Add LoadPath "CoqEAL".
Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice fintype fingroup.
Require Import finset zmodp ssralg perm rs_definitions seqmatrix1 advf_type advf_ord_triangular.
Require Import matrix seqmatrix_adds triangular triangular_rs permseq reorder 
definitions_typesCC2 Bool reduction_definition_two.
Add LoadPath "CoqEAL".

Require Import orderedcc_reduction.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Open Local Scope ring_scope.

Definition Z2 := Fp_fieldType 2. 


Section reduction_rs_CC2.



(*Chaincomplex c2 between the initial matrices*)

Record chaincomplex_advf:={
  c:> chaincomplex;
  dvf_c :seq (nat*nat);
  ords_c: orders;
  dvfnonil_c: dvf_c!=[::];
  M2nonil_c: (M2 c)!=[::];
  vecfield_c: Vecfieldadm (M1 c) dvf_c ords_c}.


Variable chaincomplexM1M2 :chaincomplex_advf.

Notation dvf:= (dvf_c chaincomplexM1M2).
Notation ords:= (ords_c chaincomplexM1M2).
Notation M1:= (M1 chaincomplexM1M2).
Notation M2:= (M2 chaincomplexM1M2).
Notation m:= (m chaincomplexM1M2).
Notation n:= (n chaincomplexM1M2).
Notation n2:= (n2 chaincomplexM1M2).

Definition s1:= fill (getfirstElemseq dvf) m.
Definition s2:= fill (getsndElemseq dvf) n.

Definition sdvf:= (size dvf).


Lemma dvfnonil: dvf!=[::].
Proof.
exact (dvfnonil_c chaincomplexM1M2).
Qed.

Lemma M2nonil: M2!=[::].
Proof.
exact (M2nonil_c chaincomplexM1M2).
Qed.

Lemma vecfield: Vecfieldadm M1 dvf ords.
Proof.
exact (vecfield_c chaincomplexM1M2).
Qed.

Lemma ismatrixM1: is_matrix m n M1.
Proof.
by move: (chaincomplex_proof chaincomplexM1M2); move =>[h _].
Qed.

Lemma ismatrixM2: is_matrix n n2 M2.
Proof.
by move: (chaincomplex_proof chaincomplexM1M2); move =>[_ [h _]].
Qed.

Lemma prodM1M20: (mx_of_seqmx m n M1) *m (mx_of_seqmx n n2 M2) = 0.
Proof.
by move: (chaincomplex_proof chaincomplexM1M2); move =>[_ [_ h]].
Qed.


Lemma sdvf_m_sdvf : (sdvf + (m - sdvf))%N = m%N.
Proof.
move: (size_dvf_reorderM dvf M1)=> H1.
rewrite /sdvf.
move: ismatrixM1; rewrite /is_matrix; move => [H0| [H3 H2]].
  by move: (dvf_nonil_M_nonil vecfield H0) dvfnonil => ->.
by move: H3; rewrite (sizeM_reorderM vecfield) => ->;rewrite(subnKC H1).
Qed.


Lemma sdvf_n_sdvf : (sdvf + (n - sdvf))%N = n%N.
Proof.
move: (size_dvf_lower_ncolumns vecfield dvfnonil ismatrixM1)=> H1.
by rewrite /sdvf (subnKC H1).
Qed.


Lemma H : perm_eq s1 (iota 0 m).
Proof.
move: vecfield; rewrite /Vecfieldadm; move => [H1 [_ [H2 _]]].
move: (perm_fill H1 H2); rewrite /s1.
by rewrite (sizeMm vecfield dvfnonil ismatrixM1).
Qed.


Lemma H1 : perm_eq s2 (iota 0 n).
Proof.
move: vecfield; rewrite /Vecfieldadm; move=> [_ [H1 [_ [ H2 _]]]].
move: (perm_fill H1 H2); rewrite /s2.
by rewrite (sizeheadMn vecfield dvfnonil ismatrixM1).
Qed.


Lemma sizes1: size s1 = m.
Proof.
by rewrite (perm_eq_size H) size_iota.
Qed.


Lemma sizes2: size s2 = n.
Proof.
by rewrite (perm_eq_size H1) size_iota.
Qed.



Section red_initialcc_orderedcc.



(*Both matrices ordered using dvf*)
Definition M1ord:= reorderM_dvf dvf M1. 
(*Order the rows of the M2 using the sequence s2, which is also used to reorder the columns of M1*)
Definition M2ord:= t_take_rows_s s2 M2.



(*Chaincomplex c1 between the ordered matrices *)
Section M1ordM2ord_chaincomplex.

Lemma ismatrix_t_take_rows: is_matrix n n2 M2ord.
Proof.
rewrite /M2ord.
move: ismatrixM2; move=>[H11|H11]; first by left; rewrite H11 /=.
rewrite -(seqmx_of_mx_mx_of_seqmx H11) t_take_rows_sE.
  by rewrite H1.
move=> H; rewrite /is_matrix; right.
move: H11=>[H2 H3].
split; first by rewrite /row_perm /= size_seqmx.
by move=> i H4; rewrite /row_perm /= /rowseqmx size_row_seqmx.
Qed.


Lemma prod_M1M2ord: (mx_of_seqmx (sdvf + (m - ( sdvf))) (sdvf + (n -sdvf)) M1ord)
   *m (mx_of_seqmx (sdvf + (n - (sdvf))) n2 M2ord) = 0.
Proof.
rewrite /M1ord /reorderM_dvf.
move: (M_nonil vecfield dvfnonil)=> H11.
move: (sizes1) => H2.
move: (sizes2) => H3.
move: (H) => H4.
move: (H1) => H5.
move: (H4); rewrite -H2 => H6.
rewrite /s1 /s2 in H2 H3 H4 H5 H6.
rewrite sizes1.
rewrite (sdvf_m_sdvf).
rewrite (sdvf_n_sdvf).
rewrite (sizeMm vecfield dvfnonil ismatrixM1) (sizeheadMn vecfield dvfnonil ismatrixM1).
rewrite -(reorderM2_E H4 H5 H11 H2 H3 H6 ismatrixM1) /M2ord.
move: (ismatrixM2); rewrite /is_matrix.
move=>[H7|H7]; first by move: M2nonil; rewrite H7.
rewrite -(seqmx_of_mx_mx_of_seqmx H7) t_take_rows_sE mx_of_seqmx_seqmx_of_mx.
rewrite /reorder_mx mul_row_perm -col_permM /s2. 
have ->:((perm_inv (perm_of_seq H5  (fill (getsndElemseq dvf) n))) * (perm_of_seq H5 (fill (getsndElemseq dvf) n)))%g = 
  (perm_mul (perm_inv (perm_of_seq H5 (fill (getsndElemseq dvf) n))) (perm_of_seq H5 (fill (getsndElemseq dvf) n)))%g; first by rewrite //.
by rewrite (perm_invP) col_perm1 row_permE -mulmxA prodM1M20 mulmx0.
Qed.


Lemma M1ordM2ord_chaincomplex: is_chaincomplex M1ord M2ord m n n2.
Proof.
rewrite /is_chaincomplex.
move: (is_matrixReorderM vecfield dvfnonil ismatrixM1 (M_nonil vecfield dvfnonil)).
move: ismatrix_t_take_rows; rewrite /M1ord /M2ord => H1 H2.
split; first by done.
split; first by done.
move: prod_M1M2ord.
by rewrite (sdvf_m_sdvf)(sdvf_n_sdvf).
Qed.

Definition chaincomplexM1M2ord := Build_chaincomplex M1ordM2ord_chaincomplex.


End M1ordM2ord_chaincomplex.


(*In this section: we build a reduction between the initial matrices and the ordered matrices*)
Section reduction_CC2_M_Mord.

(*Firstly, we will prove that there is an equivalence between both "set" matrices*)
Section equivalence_CC2_M_Mord.

Definition f1_eq := (@perm_mx Z2 m (perm_of_seq H s1)). (*change rows*)
Definition f2_eq := (@perm_mx Z2 n (perm_of_seq H1 s2 )).  (*(perm_inv (perm_of_seq H1 s2))*)(*exchange columns*)
Definition f3_eq: 'M[Z2]_n2 := (1:Z2)%:M.
Definition g1_eq := (invmx f1_eq).
Definition g2_eq := (invmx f2_eq).
Definition g3_eq :'M[Z2]_n2 := (1:Z2)%:M.




(*Chain complex morphism f (g) between the initial matrices and the ordered one 
  (between the ordered matrices and the initial one)*)
Section f_g_chaincomplex_morphism_eq.

Lemma proof_f1_M1_M2_f2: (f1_eq *m (mx_of_seqmx m n M1)) = ((mx_of_seqmx m n M1ord) *m f2_eq).
Proof.
rewrite /f1_eq /f2_eq -row_permE.
have <-: (perm_inv (perm_inv (perm_of_seq H1 s2))) =  (perm_of_seq H1 s2); first by apply: invgK.
rewrite -col_permE /M1ord /reorderM_dvf.
move: ismatrixM1; rewrite /is_matrix; move => [H2|H2].
  by move: dvfnonil; rewrite (dvf_nonil_M_nonil vecfield H2).
rewrite -{4}(seqmx_of_mx_mx_of_seqmx H2).
move: H2 =>[H3 H2]; rewrite -H3.
move: (Om vecfield dvfnonil ismatrixM1) => H5.
move: (H2 0%N H5)=> ->.
rewrite -/s1 -/s2.
move: sizes1 => H6.
move: sizes2 => H4.
move: (M_nonil_seqmx_nonil vecfield dvfnonil ismatrixM1)=> H7.
rewrite (reorderM_E H H1 H7 H6 H4) mx_of_seqmx_seqmx_of_mx /= -col_permM //.
have ->:((perm_inv (perm_of_seq H1 s2)) * (perm_of_seq H1 s2))%g = 
  (perm_mul (perm_inv (perm_of_seq H1 s2)) (perm_of_seq H1 s2))%g; first by rewrite //.
by rewrite (perm_invP) col_perm1.
Qed.



Lemma proof_f2_M'_M2_f3: f2_eq *m mx_of_seqmx n n2 M2 = mx_of_seqmx n n2 M2ord *m f3_eq.
Proof.
rewrite /f2_eq /f3_eq -row_permE mulmx1.
rewrite /M2ord.
move: ismatrixM2; rewrite /is_matrix; move => [H2|H2].
  by move: M2nonil; rewrite H2.
rewrite -{2}(seqmx_of_mx_mx_of_seqmx H2).
rewrite (t_take_rows_sE (H1)). 
by rewrite mx_of_seqmx_seqmx_of_mx.
Qed.


Lemma row_perm_inv (s:'S_n)(A:'M[Z2]_(n,n2)): (row_perm (s^-1 * s) A) = A.
Proof.
have ->: (s^-1 * s)%g = perm_mul (perm_inv s) s; first by rewrite //.
by rewrite perm_invP row_perm1.
Qed.



Lemma proof_M1_g2_g1_M2: ((mx_of_seqmx m n M1) *m g2_eq) = (g1_eq *m (mx_of_seqmx m n M1ord)).
Proof.
rewrite /g1_eq /g2_eq /f1_eq /f2_eq.
move: (On vecfield dvfnonil ismatrixM1) => H5.
have [x H2]: exists n1, n = n1.+1.
  by exists n.-1; rewrite (prednK H5).
move: H1; rewrite H2 => H0.
have <-: ((perm_mx (perm_of_seq H0 s2))^-1) = invmx (perm_mx (perm_of_seq H0 s2)); first by done.
rewrite -(perm_mxV Z2 (perm_of_seq H0 s2)) -col_permE.
move: (Om vecfield dvfnonil ismatrixM1)=> H51.
have [y H3]: exists n1, m = n1.+1.
  by exists m.-1; rewrite (prednK H51).
move: H; rewrite H3 => H4.
have <-: ((perm_mx (perm_of_seq H4 s1))^-1) = invmx (perm_mx (perm_of_seq H4 s1)); first by done.
rewrite -(perm_mxV Z2 (perm_of_seq H4 s1)) -row_permE /M1ord /reorderM_dvf.
move: ismatrixM1; rewrite /is_matrix; move => [H21|H21].
  by move: (dvf_nonil_M_nonil vecfield H21) dvfnonil => ->.
rewrite -{4}(seqmx_of_mx_mx_of_seqmx H21).
move: H21 =>[H31 H21]; rewrite -H31.
move: (H21 0%N H51)=> ->.
rewrite -/s1 -/s2 H2 H3.
move: (sizes1); rewrite H3 => H61.
move: (sizes2); rewrite H2 => H41.
move: (M_nonil_seqmx_nonil vecfield dvfnonil ismatrixM1); rewrite H3 H2 => H7.
rewrite (reorderM_E H4 H0 H7 H61 H41) mx_of_seqmx_seqmx_of_mx /= -col_row_permC -row_permM //.
have ->:((perm_inv (perm_of_seq H4 s1)) * (perm_of_seq H4 s1))%g = 
  (perm_mul (perm_inv (perm_of_seq H4 s1)) (perm_of_seq H4 s1))%g; first by rewrite //.
by rewrite (perm_invP) row_perm1.
Qed.



Lemma proof_M'_g3_g2_M2:  mx_of_seqmx n n2 M2 *m g3_eq = g2_eq *m mx_of_seqmx n n2 M2ord.
Proof.
rewrite /g2_eq  /g3_eq /f2_eq /f3_eq.
move: (On vecfield dvfnonil ismatrixM1)=> H51.
have [y H3]: exists n1, n = n1.+1.
  by exists n.-1; rewrite (prednK H51).
move: (H1); rewrite H3 => H4.
have <-: ((perm_mx (perm_of_seq H4 s2))^-1) = invmx (perm_mx (perm_of_seq H4 s2)); first by done.
rewrite -(perm_mxV Z2 (perm_of_seq H4 s2)) -row_permE /reorderM_dvf.
rewrite /M2ord.
move: ismatrixM2; rewrite /is_matrix; move => [H2|H2].
  by move: M2nonil; rewrite H2.
rewrite -{2}(seqmx_of_mx_mx_of_seqmx H2).
rewrite H3.
rewrite (t_take_rows_sE H4).
rewrite mx_of_seqmx_seqmx_of_mx.
rewrite -row_permM.
move: H4.
rewrite -H3.
move=> H4.
rewrite (row_perm_inv(perm_of_seq H4 s2)(mx_of_seqmx n n2 M2)).
by rewrite mulmx1.
Qed.


Lemma f_chaincomplex_morphism_eq: 
  (@is_chaincomplex_morphism chaincomplexM1M2 chaincomplexM1M2ord f1_eq f2_eq f3_eq).
Proof.
rewrite /is_chaincomplex_morphism /=.
split; first by rewrite proof_f1_M1_M2_f2.
by rewrite proof_f2_M'_M2_f3.
Qed.

Definition chaincomplex_morphism_f_eq := Build_chaincomplex_morphism f_chaincomplex_morphism_eq.




Lemma g_chaincomplex_morphism_eq: 
  (@is_chaincomplex_morphism chaincomplexM1M2ord chaincomplexM1M2 g1_eq g2_eq g3_eq).
Proof.
rewrite /is_chaincomplex_morphism /=.
split; first by rewrite proof_M1_g2_g1_M2.
by rewrite proof_M'_g3_g2_M2.
Qed.

Definition chaincomplex_morphism_g_eq := Build_chaincomplex_morphism g_chaincomplex_morphism_eq.


End f_g_chaincomplex_morphism_eq.


Lemma proof_f1_g1_id: f1_eq *m g1_eq = (1:Z2)%:M.
Proof.
by rewrite mulmxV; last by rewrite unitmx_perm.
Qed.

Lemma proof_g1_f1_id: g1_eq *m f1_eq = (1:Z2)%:M.
Proof.
by rewrite mulVmx; last by rewrite unitmx_perm.
Qed.


Lemma proof_f2_g2_id: f2_eq *m g2_eq = (1:Z2)%:M.
Proof.
by rewrite mulmxV; last by rewrite unitmx_perm.
Qed.

Lemma proof_g2_f2_id: g2_eq *m f2_eq = (1:Z2)%:M.
Proof.
by rewrite mulVmx; last by rewrite unitmx_perm.
Qed.


(*there is an equivalence between the initial matrices and the reordered matrices*)

Lemma is_equivalence_CC2_M_Mord: is_equivalence_CC2 chaincomplex_morphism_f_eq chaincomplex_morphism_g_eq.
Proof.
rewrite /is_equivalence_CC2.
rewrite /=.
split; first by rewrite proof_f1_g1_id.
split; first by rewrite proof_g1_f1_id.
split; first by rewrite proof_f2_g2_id.
split; first by rewrite proof_g2_f2_id.
by split; rewrite /g3_eq /f3_eq mulmx1.
Qed.


Definition equivalence_CC2_m_mord:= (Build_equivalence_CC2 is_equivalence_CC2_M_Mord).

End equivalence_CC2_M_Mord.




(*We continue proving that we can build a reduction between the initial matrices and the reordered matrices*)

Definition h1_eq: 'M[Z2]_(n,m):= 0.
Definition h2_eq: 'M[Z2]_(n2,n):= 0.

Definition homotopy_op_h_eq := (@Build_homotopy_op chaincomplexM1M2 h1_eq h2_eq).


(*there is an reduction between the initial matrices and the reordered matrices*)
Lemma is_reduction_CC2_eq: is_reduction_CC2 chaincomplex_morphism_f_eq chaincomplex_morphism_g_eq  homotopy_op_h_eq.
Proof.
rewrite /is_reduction_CC2 /=.
move: ( is_equivalence_CC2_M_Mord); rewrite /is_equivalence_CC2 /=.
rewrite /h2_eq /h1_eq.
move => [H1 [H2 [H3 [H4 [H5 H6]]]]].
split; first by done.
split; first by done.
split; first by done.
split; rewrite /h1_eq mulmx0.
  have ->: (g1_eq *m f1_eq + 0 = (addmx (g1_eq *m f1_eq) 0)); first by rewrite //.
  by rewrite addmxC add0mx.
split. rewrite mul0mx.
  have ->: (g2_eq *m f2_eq + 0 + 0 = (addmx (addmx (g2_eq *m f2_eq) 0) 0)); first by rewrite //.
  by rewrite addmxC add0mx addmxC add0mx.
split.
 rewrite mul0mx.
  have ->: (g3_eq *m f3_eq + 0 = (addmx (g3_eq *m f3_eq) 0)); first by rewrite //.
  by rewrite addmxC add0mx.
split; first by rewrite mulmx0.
split; first by rewrite mulmx0.
split; first by rewrite mul0mx.
split; first by rewrite mul0mx.
by rewrite mul0mx.
Qed.

(*reduction from C to the ordered *)
Definition reduct_eq := (Build_reduction_CC2 is_reduction_CC2_eq).

End reduction_CC2_M_Mord. 

End red_initialcc_orderedcc.



(*-------------------------------- *)


(*Reduction applying BPL*)

Section red_ord_BPL.


(*Due to the kernel, we have to define the differentials with their transposes*)
Definition M1ord_trmx := 
  trmx ((mx_of_seqmx (sdvf + (m - ( sdvf))) (sdvf + (n -sdvf)) M1ord)).

Definition M2ord_trmx :=  trmx (mx_of_seqmx (sdvf + (n - (sdvf))) n2 M2ord).


(*Properties to be able to apply BPL*)
Lemma prodM1M2ord_trmx : M2ord_trmx *m M1ord_trmx = 0.
Proof.
rewrite /M2ord_trmx /M1ord_trmx.
by rewrite -trmx_mul prod_M1M2ord trmx0.
Qed.


Lemma daa_triangular_upper : triangular_upper.upper_triangular (daa M1ord_trmx).
Proof. 
rewrite /M1ord_trmx /daa.
rewrite -trmx_ulsub.
apply:  triangular_upper.lower_triangular_trmx.
rewrite /M1ord.
move: (lower_triangular_ulsubseqmx_rs dvfnonil (vecfield_c chaincomplexM1M2)).
rewrite -(submxK ((mx_of_seqmx
           (sdvf + (definitions_typesCC2.m chaincomplexM1M2 - sdvf))
           (sdvf + (definitions_typesCC2.n chaincomplexM1M2 - sdvf))
           (reorderM_dvf dvf (definitions_typesCC2.M1 chaincomplexM1M2))))).
rewrite block_mxKul.
rewrite -(mx_of_seqmx_seqmx_of_mx (ulsubmx (mx_of_seqmx
           (sdvf + (definitions_typesCC2.m chaincomplexM1M2 - sdvf))
           (sdvf + (definitions_typesCC2.n chaincomplexM1M2 - sdvf))
           (reorderM_dvf dvf (definitions_typesCC2.M1 chaincomplexM1M2))))).
rewrite -ulsubseqmxE.
rewrite seqmx_of_mx_mx_of_seqmx /=; last first.
  move: (is_matrixReorderM vecfield dvfnonil ismatrixM1 (M_nonil vecfield dvfnonil)).
  rewrite /is_matrix /=.
  move=>  [H1| H2]; last first.
    by move:H2; rewrite sdvf_m_sdvf sdvf_n_sdvf => HH.
have H: (size (reorderM_dvf dvf (definitions_typesCC2.M1 chaincomplexM1M2)) = 0%N).
  by rewrite H1.
move: (size_dvf_reorderM dvf (definitions_typesCC2.M1 chaincomplexM1M2)) .
  rewrite H.
  rewrite leqn0; move/eqP=> HH.
  move: dvfnonil.
  by rewrite (size0nil HH).
rewrite /sdvf.
case H: (size dvf) =>[|n].
  by move: (dvfnonil); rewrite (size0nil H).
rewrite !subn1.
rewrite NPeano.Nat.pred_succ.
rewrite /lower_triangular /triangular.lower_triangular.
by rewrite addn1.
Qed. 


Require Import definitions_auxlemmas.


(*We define a general reduction between the chain complex composed by the transpose
  differentials C and another one D reduced applying the BPL*)
Definition red_BPL:= reduction_BPL_general prodM1M2ord_trmx daa_triangular_upper.






(*---------------------------------*)

(* - Reduction CC2
   - the initial matrices of this new reduction have to be the ones of the ordered 
     chain complex *)

(*Define a reduction from red_BPL composed by the no null differentials, deleting the null differentials where the differentials of the initial chain complex are the tranpose of the differentials of C*)


Section redBPL_redCC2_transpose.

Definition diff0C:= seqmx_of_mx (trmx (diff (C red_BPL) 0)).
Definition diff1C:= seqmx_of_mx (trmx (diff (C red_BPL) 1)).


Definition mC:= (m (C red_BPL) 0).
Definition nC:= (m (C red_BPL) 1).
Definition n2C:= (m (C red_BPL) 2).


(*Convert (D red_BPL) in a chain complex*)

Definition diff0D:= seqmx_of_mx (trmx (diff (D red_BPL) 0)).
Definition diff1D:= seqmx_of_mx (trmx (diff (D red_BPL) 1)).

Definition mD:= (m (D red_BPL) 0).
Definition nD:= (m (D red_BPL) 1).
Definition n2D:= (m (D red_BPL) 2).


(*relations between the differentials of C with M1ord and M2ord and theirs tranposes*)
Lemma diffC0_M1T : (diff (C red_BPL) 0) = M1ord_trmx.
Proof.
rewrite /(diff (C red_BPL) 0) /= /BPL.Di_def /= /BPL.Di_BPL //= /P1.
rewrite GRing.addrA [(delta1 Z2 sdvf (definitions_typesCC2.n chaincomplexM1M2 - sdvf)
      (definitions_typesCC2.m chaincomplexM1M2 - sdvf)) + _]GRing.addrC -GRing.addrA.
rewrite m_m_0 GRing.addr0.
by rewrite castmx_id.
Qed.


Lemma diffC1_M2T :(diff (C red_BPL) 1) = M2ord_trmx.
Proof.
rewrite /(diff (C red_BPL) 1) /= /BPL.Di_def /= /BPL.Di_BPL //= /P2.
rewrite /redCC2_general.D2 GRing.add0r.
by rewrite castmx_id.
Qed.



(*diff0D and diff1D composed a 2-dimensional chain complex*)

Lemma ismatrix_D1_seq :  is_matrix mD nD diff0D. 
Proof.
rewrite /diff0D.
rewrite /is_matrix.
 right.
split; first by rewrite size_seqmx.
move=> i H.
by rewrite size_row_seqmx.
Qed.

Lemma ismatrix_D2_seq :  is_matrix nD n2D diff1D. 
Proof.
rewrite /diff1D.
rewrite /is_matrix.
 right.
split; first by rewrite size_seqmx.
move=> i H.
by rewrite size_row_seqmx.
Qed.

Lemma D_prod:  mx_of_seqmx mD nD diff0D *m mx_of_seqmx nD n2D diff1D = 0.
Proof.
move: (nilpotency (D red_BPL) 0).
rewrite /diff0D /diff1D.
rewrite !mx_of_seqmx_seqmx_of_mx -trmx_mul => ->.
by rewrite trmx0.
Qed.

Lemma is_chaincomplexD : is_chaincomplex diff0D diff1D mD nD n2D.
Proof.
rewrite /is_chaincomplex.
split; first by apply: ismatrix_D1_seq.
split; first by apply: ismatrix_D2_seq.
by rewrite D_prod.
Qed.


Definition ccD := Build_chaincomplex is_chaincomplexD.




(*Auxiliar lemmas to use casts*)

Lemma a0: m (D red_BPL) 0 = (definitions_typesCC2.m ccD).
Proof.
trivial.
Qed.

Lemma b0: m (C red_BPL) 0 = (definitions_typesCC2.m chaincomplexM1M2ord).
Proof.
by rewrite /= sdvf_m_sdvf.
Qed.

Lemma b00: (definitions_typesCC2.m chaincomplexM1M2ord) =  m (C red_BPL) 0.
Proof.
by rewrite /= sdvf_m_sdvf.
Qed.

Lemma a1:m (D red_BPL) 1 = (definitions_typesCC2.n ccD).
Proof.
trivial.
Qed.

Lemma b1: m (C red_BPL) 1 = (definitions_typesCC2.n chaincomplexM1M2ord).
Proof.
by rewrite /= sdvf_n_sdvf.
Qed.

Lemma b10: (definitions_typesCC2.n chaincomplexM1M2ord) =  m (C red_BPL) 1.
Proof.
by rewrite /= sdvf_n_sdvf.
Qed.

Lemma a2:m (D red_BPL) 2 = (definitions_typesCC2.n2 ccD).
Proof.
trivial.
Qed.

Lemma b2: m (C red_BPL) 2 = (definitions_typesCC2.n2 chaincomplexM1M2ord).
Proof.
trivial.
Qed.




(*Auxiliar lemmas about the transpose matrix*)

Lemma trmx_trmx: forall (R : comRingType) (n1 n2 : nat) (A : 'M[R]_(n1,n2)),
   (A^T)^T = A.
Proof.
move=> R n1 n2 A. 
by rewrite trmxK.
Qed.

Lemma add_trmx : forall (R : ringType) (m1 n1 m2 n2 : nat) 
     (A : 'M[R]_(m1, n1)) (B :'M[R]_(m1,n1)), 
A^T + B^T = (A + B)^T.
Proof.
move=> R m1 n1 m2 n2 A B.
by apply/matrixP=> k i; rewrite !mxE.
Qed.



(* Auxiliar lemmas
relations between the differentials of the chain complex general reduced (D red_BPL)
 and the new chain complex (2matrices) reduced*)
Lemma M1C :  (mx_of_seqmx (definitions_typesCC2.m chaincomplexM1M2)
     (definitions_typesCC2.n chaincomplexM1M2) M1ord) =
     (castmx (b0,b1)  (trmx (diff (C red_BPL) 0))).
Proof.
by rewrite diffC0_M1T /M1ord_trmx trmx_trmx cast_mx_of_seqmx.
Qed.

Lemma M2C :  (mx_of_seqmx (definitions_typesCC2.n chaincomplexM1M2)
     (definitions_typesCC2.n2 chaincomplexM1M2) M2ord) =
     (castmx (b1,b2)  (trmx (diff (C red_BPL) 1))).
Proof.
by rewrite diffC1_M2T /M2ord_trmx trmx_trmx cast_mx_of_seqmx.
Qed.

Lemma M1D :  (mx_of_seqmx (definitions_typesCC2.m ccD)
     (definitions_typesCC2.n ccD) (definitions_typesCC2.M1 ccD)) =
      (trmx (diff (D red_BPL) 0)).
Proof.
rewrite /red_BPL /=.
by rewrite /diff0D /red_BPL /= mx_of_seqmx_seqmx_of_mx.
Qed.

Lemma M2D :  (mx_of_seqmx (definitions_typesCC2.n ccD)
     (definitions_typesCC2.n2 ccD) (definitions_typesCC2.M2 ccD)) =
     (trmx (diff (D red_BPL) 1)).
Proof.
rewrite /red_BPL /=.
by rewrite /diff1D /red_BPL /= mx_of_seqmx_seqmx_of_mx.
Qed.



(*We define the morphism F (cc 2 dimensional) from the general morphism (F red_BPL)
 which comes from the reduction obtained after applying BPL *)

Definition F0:=  castmx (a0,b0) (trmx (M (F red_BPL) 0)).
Definition F1:= castmx (a1,b1) (trmx (M (F red_BPL) 1)).
Definition F2:= castmx (a2,b2) (trmx (M (F red_BPL) 2)).


Lemma F0D_DF1: F0 *m mx_of_seqmx (definitions_typesCC2.m chaincomplexM1M2ord)
           (definitions_typesCC2.n chaincomplexM1M2ord)
           (definitions_typesCC2.M1 chaincomplexM1M2ord) =
   mx_of_seqmx (definitions_typesCC2.m ccD) (definitions_typesCC2.n ccD)
     (definitions_typesCC2.M1 ccD) *m F1.
Proof.
move: (M_well_defined (F red_BPL) 0)  => HH.
rewrite /F0 /F1 M1C M1D.
rewrite -(castmx_id (a0,a1) ((diff (D red_BPL) 0)^T)).
rewrite (mulmx_cast (a0,b0)(b0,b1)). split; first by rewrite a0. 
by rewrite b1. 
rewrite (mulmx_cast (a0,a1)(a1,b1)). split; first by rewrite a0. 
by rewrite b1. 
move=> H0 H1.
rewrite -!trmx_mul HH.
have ->: H0 = H1; last by done.
move: H0; move =>[H4 H5].
move: H1; move => [H6 H7].
rewrite (eq_irrelevance H4 H6).
by rewrite (eq_irrelevance H5 H7).
Qed.

Lemma F1D_DF2:  F1 *m mx_of_seqmx (definitions_typesCC2.n chaincomplexM1M2ord)
           (definitions_typesCC2.n2 chaincomplexM1M2ord)
           (definitions_typesCC2.M2 chaincomplexM1M2ord) =
   mx_of_seqmx (definitions_typesCC2.n ccD) (definitions_typesCC2.n2 ccD)
     (definitions_typesCC2.M2 ccD) *m F2.
Proof.
move: (M_well_defined (F red_BPL) 1)  => HH.
rewrite /F2 /F1 M2C M2D.
rewrite -(castmx_id (a1,a2) ((diff (D red_BPL) 1)^T)).
rewrite (mulmx_cast (a1,b1)(b1,b2)). split; first by rewrite a1. 
by rewrite b2. 
rewrite (mulmx_cast (a1,a2)(a2,b2)). split; first by rewrite a1. 
by rewrite b2. 
move=> H0 H1.
rewrite -!trmx_mul HH.
have ->: H0 = H1; last by done.
move: H0; move =>[H4 H5].
move: H1; move => [H6 H7].
rewrite (eq_irrelevance H4 H6).
by rewrite (eq_irrelevance H5 H7).
Qed.

Lemma is_chaincomplex_morphismF: (@is_chaincomplex_morphism chaincomplexM1M2ord ccD F0 F1 F2).
Proof.
rewrite /is_chaincomplex_morphism.
split; first by apply:F0D_DF1.
by apply: F1D_DF2.
Qed.

Definition ccmF:= Build_chaincomplex_morphism is_chaincomplex_morphismF.
  

(*morphism G*)

Definition G0:= castmx (b0,a0) (trmx (M (G red_BPL) 0)).
Definition G1:= castmx (b1,a1) (trmx (M (G red_BPL) 1)).
Definition G2:= castmx (b2,a2) (trmx (M (G red_BPL) 2)).


Lemma G0D_DG1: G0 *m mx_of_seqmx (definitions_typesCC2.m ccD)
           (definitions_typesCC2.n ccD) (definitions_typesCC2.M1 ccD) =
   mx_of_seqmx (definitions_typesCC2.m chaincomplexM1M2ord)
     (definitions_typesCC2.n chaincomplexM1M2ord)
     (definitions_typesCC2.M1 chaincomplexM1M2ord) *m G1.
Proof.
move: (M_well_defined (G red_BPL) 0)  => HH.
rewrite /G0 /G1 M1D M1C.
rewrite -(castmx_id (a0,a1) ((diff (D red_BPL) 0)^T)).
rewrite (mulmx_cast (b0,a0)(a0,a1)). split; first by rewrite b0. 
by rewrite a1. 
rewrite (mulmx_cast (b0,b1)(b1,a1)). split; first by rewrite b0. 
by rewrite a1. 
move=> H0 H1.
rewrite -!trmx_mul HH.
have ->: H0 = H1; last by done.
move: H0; move =>[H4 H5].
move: H1; move => [H6 H7].
rewrite (eq_irrelevance H4 H6).
by rewrite (eq_irrelevance H5 H7).
Qed.

Lemma G1D_DG2:
   G1 *m mx_of_seqmx (definitions_typesCC2.n ccD)
           (definitions_typesCC2.n2 ccD) (definitions_typesCC2.M2 ccD) =
   mx_of_seqmx (definitions_typesCC2.n chaincomplexM1M2ord)
     (definitions_typesCC2.n2 chaincomplexM1M2ord)
     (definitions_typesCC2.M2 chaincomplexM1M2ord) *m G2.
Proof.
move: (M_well_defined (G red_BPL) 1)  => HH.
rewrite /G2 /G1 M2C M2D.
rewrite -(castmx_id (a1,a2) ((diff (D red_BPL) 1)^T)).
rewrite (mulmx_cast (b1,a1)(a1,a2)). split; first by rewrite b1. 
by rewrite a2. 
rewrite (mulmx_cast (b1,b2)(b2,a2)). split; first by rewrite b1. 
by rewrite a2. 
move=> H0 H1.
rewrite -!trmx_mul HH.
have ->: H0 = H1; last by done.
move: H0; move =>[H4 H5].
move: H1; move => [H6 H7].
rewrite (eq_irrelevance H4 H6).
by rewrite (eq_irrelevance H5 H7).
Qed.

Lemma is_chaincomplex_morphismG: (@is_chaincomplex_morphism ccD chaincomplexM1M2ord G0 G1 G2).
Proof.
rewrite /is_chaincomplex_morphism.
split; first by apply:G0D_DG1.
by apply: G1D_DG2.
Qed.

Definition ccmG:= Build_chaincomplex_morphism is_chaincomplex_morphismG.



(*Homotopy operator (cc 2 dimensional)*)

Definition Ho0 := castmx (b1,b0) (trmx (Ho (H red_BPL) 0)).
Definition Ho1 := castmx (b2,b1) (trmx (Ho (H red_BPL) 1)).

Definition homotopy_opC:= (@Build_homotopy_op chaincomplexM1M2ord Ho0 Ho1).



(*Properties of a reduction of a chain complex 2 dimensional*)
Lemma f1g1_ccmFG: p1 ccmF *m p1 ccmG = 1%:M.
Proof.
move: (ax1 red_BPL 0) => h0.
rewrite /=.
rewrite /F0 /G0.
rewrite mulmx_cast. split; try by rewrite /=.
move=> HH.
rewrite -trmx_mul h0.
by rewrite trmx1 castmx_id.
Qed.

Lemma f2g2_ccmFG: p2 ccmF *m p2 ccmG = 1%:M.
Proof.
move: (ax1 red_BPL 1) => h0.
rewrite /=.
rewrite /F1 /G1.
rewrite mulmx_cast. split; try by rewrite /=.
move=> HH.
rewrite -trmx_mul h0.
by rewrite trmx1 castmx_id.
Qed.

Lemma f3g3_ccmFG: p3 ccmF *m p3 ccmG = 1%:M.
Proof.
move: (ax1 red_BPL 2) => h0.
rewrite /=.
rewrite /F2 /G2.
rewrite mulmx_cast. split; try by rewrite /=.
move=> HH.
rewrite -trmx_mul h0.
by rewrite trmx1 castmx_id.
Qed.




Lemma g1f1_dh1_ccmFG:  p1 ccmG *m p1 ccmF +
   mx_of_seqmx (definitions_typesCC2.m chaincomplexM1M2ord)
     (definitions_typesCC2.n chaincomplexM1M2ord)
     (definitions_typesCC2.M1 chaincomplexM1M2ord) *m 
   h1 homotopy_opC = 1%:M.
Proof.
move: (ax2 red_BPL (-1)) => h0.
rewrite /= /G0 /F0 /Ho0.
move: h0.
have ->: (diff (C red_BPL) (-1)) = 0.
  rewrite /red_BPL /= /BPL.Di_def /BPL.Di_BPL //=.
  by rewrite !thinmx0.
rewrite mul0mx GRing.addr0.
move=> h0.
rewrite M1C.
rewrite !mulmx_cast. split; try by rewrite /=.
by rewrite /= sdvf_m_sdvf.
by rewrite /= sdvf_m_sdvf.
split; try by rewrite /= sdvf_m_sdvf.
move=> HH HH1.
rewrite -!trmx_mul.
have ->: HH = HH1.
  move: HH; move =>[H4 H5].
  move: HH1; move => [H6 H7].
  rewrite (eq_irrelevance H4 H6).
  by rewrite (eq_irrelevance H5 H7).
rewrite BPL.addmx_cast.
rewrite (add_trmx (m (C red_BPL) 0) (m (C red_BPL) 0)). 
by rewrite h0 trmx1 BPL.id_castmx.
Qed.

Lemma g2f2_dh2_h1d_ccmFG: p2 ccmG *m p2 ccmF +
   mx_of_seqmx (definitions_typesCC2.n chaincomplexM1M2ord)
     (definitions_typesCC2.n2 chaincomplexM1M2ord)
     (definitions_typesCC2.M2 chaincomplexM1M2ord) *m 
   h2 homotopy_opC +
   h1 homotopy_opC *m mx_of_seqmx
                        (definitions_typesCC2.m chaincomplexM1M2ord)
                        (definitions_typesCC2.n chaincomplexM1M2ord)
                        (definitions_typesCC2.M1 chaincomplexM1M2ord) = 1%:M.
Proof.
move: (ax2 red_BPL 0) => h0.
rewrite /= /G1 /F1 /Ho0 /Ho1.
rewrite M1C M2C.
rewrite !mulmx_cast. split; try by rewrite /=.
by rewrite /= sdvf_n_sdvf.
by rewrite /= sdvf_n_sdvf.
split; try by rewrite /= sdvf_n_sdvf.
split; try by rewrite /= sdvf_n_sdvf.
move=> HH HH1 HH2.
rewrite -!trmx_mul.
have ->: HH = HH1.
  move: HH; move =>[H4 H5].
  move: HH1; move => [H6 H7].
  rewrite (eq_irrelevance H4 H6).
  by rewrite (eq_irrelevance H5 H7).
have ->: HH2 = HH1.
  move: HH2; move =>[H4 H5].
  move: HH1; move => [H6 H7].
  rewrite (eq_irrelevance H4 H6).
  by rewrite (eq_irrelevance H5 H7).
rewrite !BPL.addmx_cast.
rewrite !(add_trmx (m (C red_BPL) 0) (m (C red_BPL) 0)). 
by rewrite h0 trmx1 BPL.id_castmx.
Qed.

Lemma g3f3_h2d_ccmFG:   p3 ccmG *m p3 ccmF +
   h2 homotopy_opC *m mx_of_seqmx
                        (definitions_typesCC2.n chaincomplexM1M2ord)
                        (definitions_typesCC2.n2 chaincomplexM1M2ord)
                        (definitions_typesCC2.M2 chaincomplexM1M2ord) = 1%:M.
Proof.
move: (ax2 red_BPL 1) => h0.
rewrite /= /G2 /F2 /Ho1.
move:h0.
have ->: ( diff (C red_BPL)  (BinInt.Z.add 1 1)) = 0.
  rewrite /red_BPL /= /BPL.Di_def /BPL.Di_BPL //=.
  by rewrite !flatmx0. 
rewrite mulmx0 GRing.addr0 => h0.
rewrite M2C.
rewrite !mulmx_cast. split; try by rewrite /=.
split; try by rewrite /=.
move=> HH HH1.
rewrite -!trmx_mul.
have ->: HH = HH1.
  move: HH; move =>[H4 H5].
  move: HH1; move => [H6 H7].
  rewrite (eq_irrelevance H4 H6).
  by rewrite (eq_irrelevance H5 H7).
rewrite BPL.addmx_cast.
rewrite (add_trmx (m (C red_BPL) 0) (m (C red_BPL) 0)). 
by rewrite h0 trmx1 BPL.id_castmx.
Qed.



Lemma f2h1_ccmFG:  p2 ccmF *m h1 homotopy_opC = 0.
Proof.
move: (ax3 red_BPL 0) => h0.
rewrite /=.
rewrite /F1 /Ho0.
rewrite mulmx_cast. split; try by rewrite /=.
  by rewrite /= sdvf_m_sdvf.
move=> HH.
rewrite -trmx_mul h0.
rewrite trmx0.
by rewrite castmx_const.
Qed.

Lemma f3h2_ccmFG:  p3 ccmF *m h2 homotopy_opC = 0.
Proof.
move: (ax3 red_BPL 1) => h0.
rewrite /=.
rewrite /F2 /Ho1.
rewrite mulmx_cast. split; try by rewrite /=.
  by rewrite /= sdvf_n_sdvf.
move=> HH.
rewrite -trmx_mul h0.
rewrite trmx0.
by rewrite castmx_const.
Qed.



Lemma h1g1_ccmFG:  h1 homotopy_opC *m p1 ccmG = 0.
Proof.
move: (ax4 red_BPL 0) => h0.
rewrite /=.
rewrite /G0 /Ho0.
rewrite mulmx_cast. split; try by rewrite /=.
  by rewrite /= sdvf_n_sdvf.
move=> HH.
rewrite -trmx_mul h0.
rewrite trmx0.
by rewrite castmx_const.
Qed.

Lemma h2g2_ccmFG:  h2 homotopy_opC *m p2 ccmG = 0.
Proof.
move: (ax4 red_BPL 1) => h0.
rewrite /=.
rewrite /G1 /Ho1.
rewrite mulmx_cast. split; try by rewrite /=.
move=> HH.
rewrite -trmx_mul h0.
rewrite trmx0.
by rewrite castmx_const.
Qed.



Lemma h2h1_ccmFG: h2 homotopy_opC *m h1 homotopy_opC = 0.
Proof.
move: (ax5 red_BPL 0) => h0.
rewrite /=.
rewrite /Ho0 /Ho1.
rewrite mulmx_cast. split; try by rewrite /=.
  by rewrite /= sdvf_m_sdvf.
move=> HH.
rewrite -trmx_mul h0 trmx0.
by rewrite castmx_const.
Qed.



Lemma is_r: is_reduction_CC2 ccmF ccmG homotopy_opC.
Proof.
rewrite /is_reduction_CC2.
split; first by rewrite f1g1_ccmFG.
split; first by rewrite f2g2_ccmFG.
split; first by rewrite f3g3_ccmFG.
split; first by rewrite g1f1_dh1_ccmFG.
split; first by rewrite g2f2_dh2_h1d_ccmFG.
split; first by rewrite g3f3_h2d_ccmFG.
split; first by rewrite f2h1_ccmFG.
split; first by rewrite f3h2_ccmFG.
split; first by rewrite h1g1_ccmFG.
split; first by rewrite h2g2_ccmFG.
by rewrite h2h1_ccmFG.
Qed.


Definition red_BPL_CC2:= Build_reduction_CC2 is_r.


End redBPL_redCC2_transpose.


(*Composition of reduct_eq: C -> Cord
                 red_BPL_CC2: Cord -> Cred*)
Definition red_M1M2_rs := Reduction_red_red reduct_eq red_BPL_CC2.


End red_ord_BPL.


End reduction_rs_CC2.



