Add LoadPath "CoqEAL".
Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop  ZArith Zinfra. 
Require Import definitions_auxlemmas BPL.
(*Require Import reduction_definition_two.*)

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section BPL.

Open Local Scope ring_scope.


(*In order to apply the BPL, firstly, we need a reduction. In this case this
  reduction will be special because D2, D2', F2, G2 and H1 are determined.
  Some of the properties of the reduction are trivial due to the concrete 
  definition of some of these matrices and other ones will be hypotheses.

  In this way, we build a reduction which we will use to apply BPL
*)

Section decomposition.

Variable K : fieldType.
Variables (c0 c1 c2 d0 d1 : nat).
Variables (D1 : 'M[K]_(c1,c0)) (D1' : 'M[K]_(d1,d0))
          (F0 : 'M[K]_(c0,d0)) (F1 : 'M[K]_(c1,d1)) 
          (G0 : 'M[K]_(d0,c0)) (G1 : 'M[K]_(d1,c1)) 
          (H0 : 'M[K]_(c0,c1)).

Definition D2 := 0 : 'M[K]_(c2,c1).
Definition D2' := 0 : 'M[K]_(c2,d1).
Definition F2 := 1%:M : 'M[K]_(c2,c2).
Definition G2 := 1%:M : 'M[K]_(c2,c2).
Definition H1 := 0 : 'M[K]_(c1,c2).


(*Definition chain complex D*)

Definition m_m: Z -> nat:= fun i:Z => match i with
|0%Z => c0
|1%Z => c1
|2%Z => c2
|_ => 0%nat
end.


Variable (D3 : 'M[K]_(0,c2)).
Variable (D0 : 'M[K]_(c0,0)).
Variable (Dn : 'M[K]_(0,0)).


Definition diff_m:= fun i:Z => match i as
 z return ((fun z0 : Z => 'M_(m_m (z0 + 1), m_m z0)) z) with
|(-1)%Z => D0
|0%Z => D1
|1%Z => D2
|2%Z => D3
|3%Z => Dn
|_ => Dn
end.


Lemma D2D1: D2 *m D1 = 0.
Proof.
by rewrite mul0mx.
Qed.

Lemma DnDn: Dn *m Dn = 0.
Proof.
by rewrite thinmx0.
Qed.

Lemma D3D2: D3 *m D2 = 0.
Proof.
by rewrite mulmx0.
Qed.


Lemma D1D0: D1 *m D0 = 0.
Proof.
by rewrite thinmx0.
Qed.


Lemma D0Dn: D0 *m Dn = 0.
Proof.
by rewrite thinmx0.
Qed.

Lemma DnD3: Dn *m D3 = 0.
Proof.
by rewrite flatmx0.
Qed.


Lemma nilpotency_diff_m: forall i:Z, (diff_m (i+1)) *m (diff_m i) = 0.
Proof.
move=> i.
case: i=> [|p| p].
  rewrite /=.
  exact D2D1.
case:p.
  move=> p.
  case:p; try by move=> p; rewrite /=; exact DnDn.
  rewrite /=; exact DnDn.
  move=> p.
  case:p; try by move=> p; rewrite /=; exact DnDn.
  rewrite /=; exact DnD3.
  rewrite /=; exact D3D2.
case: p; rewrite /=.
  move=> p; exact DnDn.
move=> p; rewrite //=.
case: ( Pos.pred_double p); try by move=> p0; exact DnDn.
exact D0Dn.
exact D1D0.
Qed.


Definition Dcc:= Build_FGChain_Complex nilpotency_diff_m.



(*Definition chain complex D'*)


Definition m_m_d: Z -> nat:= fun i:Z => match i with
|0%Z => d0
|1%Z => d1
|2%Z => c2
|_ => 0%nat
end.


Variable (D3' : 'M[K]_(0,c2)).
Variable (D0' : 'M[K]_(d0,0)).
Variable (Dn' : 'M[K]_(0,0)).


Definition diff_m_d:= fun i:Z => match i as
 z return ((fun z0 : Z => 'M_(m_m_d (z0 + 1), m_m_d z0)) z) with
|(-1)%Z => D0'
|0%Z => D1'
|1%Z => D2'
|2%Z => D3'
|3%Z => Dn'
|_ => Dn'
end.

Lemma D20': D2' = 0.
Proof.
by done.
Qed.

Lemma D2D1': D2' *m D1' = 0.
Proof.
by rewrite D20' mul0mx.
Qed.

Lemma DnDn': Dn' *m Dn' = 0.
Proof.
by rewrite thinmx0.
Qed.

Lemma D3D2': D3' *m D2' = 0.
Proof.
by rewrite D20' mulmx0.
Qed.


Lemma D1D0': D1' *m D0' = 0.
Proof.
by rewrite thinmx0.
Qed.


Lemma D0Dn': D0' *m Dn' = 0.
Proof.
by rewrite thinmx0.
Qed.

Lemma DnD3': Dn' *m D3' = 0.
Proof.
by rewrite flatmx0.
Qed.


Lemma nilpotency_diff_m_d: forall i:Z, (diff_m_d (i+1)) *m (diff_m_d i) = 0.
Proof.
move=> i.
case: i=> [|p| p].
  rewrite /=.
  exact D2D1'.
case:p.
  move=> p.
  case:p; try by move=> p; rewrite /=; exact DnDn'.
  rewrite /=; exact DnDn'.
  move=> p.
  case:p; try by move=> p; rewrite /=; exact DnDn'.
  rewrite /=; exact DnD3'.
  rewrite /=; exact D3D2'.
case: p; rewrite /=.
  move=> p; exact DnDn'.
move=> p; rewrite //=.
case: ( Pos.pred_double p); try by move=> p0; exact DnDn'.
exact D0Dn'.
exact D1D0'.
Qed.


Definition D'cc:= Build_FGChain_Complex nilpotency_diff_m_d.






Hypothesis F0_morph : (D1 *m F0) = (F1 *m D1').
Hypothesis G0_morph : (D1' *m G0) = (G1 *m D1).
Hypothesis ax1_0 : (G0 *m F0) = 1%:M.
Hypothesis ax1_1 : (G1 *m F1) = 1%:M.
Hypothesis ax2_0 : (F0 *m G0) + (H0 *m D1) = 1%:M.
Hypothesis ax2_1 : (F1 *m G1)  + (H1 *m D2) + (D1 *m H0) = 1%:M.
Hypothesis ax3_0 : (H0 *m F1) = 0.
Hypothesis ax4_0 : (G0 *m H0) = 0.

Lemma nilpotency_d2d1 : (D2 *m D1) = 0. 
Proof. by rewrite mul0mx. Qed.

Lemma nilpotency_d2d1' : (D2' *m D1') = 0.
Proof. by rewrite mul0mx. Qed.

Lemma F1_morph : (D2 *m F1) = (F2 *m D2').
Proof. by rewrite mul0mx mulmx0. Qed. 

Lemma G1_morph : (D2' *m G1) = (G2 *m D2).
Proof. by rewrite mul0mx mulmx0. Qed. 

Lemma ax1_2 : (G2 *m F2) = 1%:M. 
Proof. by rewrite mul1mx. Qed.

Lemma ax2_2 : (F2 *m G2) + (D2 *m H1) = 1%:M.
Proof. by rewrite mul1mx mul0mx GRing.addr0. Qed.

Lemma ax3_1 : (H1 *m F2) = 0.
Proof. by rewrite mul0mx. Qed.

Lemma ax4_1 : (G1 *m H1) = 0.
Proof. by rewrite mulmx0. Qed. 

Lemma ax5 : (H0 *m H1) = 0.
Proof. by rewrite mulmx0. Qed. 




(*Definition F_m and G_m chain complex morphisms*)

Variable (Fn : 'M[K]_(0,0)).

Definition F_m:= fun i:Z => match i as
 z return ((fun z0 : Z => 'M_(m_m z0 , m_m_d z0)) z) with
(*|(-1)%Z => D0'*)
|0%Z => F0
|1%Z => F1
|2%Z => F2
|_ => Fn
end.


Lemma D1F0F1D1' : D1 *m F0 = F1 *m D1'.
Proof.
by apply: F0_morph.
Qed.

Lemma D2F1F2D2' : D2 *m F1 = F2 *m D2'.
Proof.
by apply: F1_morph.
Qed.

Lemma D3F2FnD3' : D3 *m F2 = Fn *m D3'.
Proof.
by rewrite thinmx0 flatmx0 mul0mx.
Qed.

Lemma DnFnFnDn': Dn *m Fn = Fn *m Dn'.
Proof.
by rewrite thinmx0 flatmx0.
Qed.


Lemma D0FnF0D0':D0 *m Fn = F0 *m D0'.
Proof.
by rewrite !thinmx0.
Qed.



Lemma F_ccmorphism_m:
   forall i:Z, (diff Dcc i) *m (F_m i) = (F_m (i+1)) *m (diff D'cc i).
Proof.
move=> i.
case:i.
  rewrite /=; exact D1F0F1D1'.
move=> p.
case :p.
  move=> p.
  case:p; try by move=>p; rewrite /=; exact DnFnFnDn'.
  rewrite /=; exact DnFnFnDn'.
  move=>p.
  case:p; try by move=> p; rewrite /=; exact DnFnFnDn'.
  rewrite /=; exact D3F2FnD3'.
  rewrite /=; exact  D2F1F2D2'.
move=> p.
case:p; try by move=>p; rewrite /=; exact DnFnFnDn'.
rewrite /=; exact D0FnF0D0'. 
Qed.


Definition F_m_m:= Build_FGChain_Complex_Morphism (A:=Dcc)(B:=D'cc) F_ccmorphism_m.




Variable (Gn : 'M[K]_(0,0)).

Definition G_m:= fun i:Z => match i as
 z return ((fun z0 : Z => 'M_(m_m_d z0 , m_m z0)) z) with
|0%Z => G0
|1%Z => G1
|2%Z => G2
|_ => Gn
end.


Lemma D1G0G1D1' : D1' *m G0 = G1 *m D1.
Proof.
by apply: G0_morph.
Qed.

Lemma D2G1G2D2' : D2' *m G1 = G2 *m D2.
Proof.
by apply: G1_morph.
Qed.

Lemma D3G2GnD3' : D3' *m G2 = Gn *m D3.
Proof.
by rewrite thinmx0 flatmx0 mul0mx.
Qed.

Lemma DnGnGnDn': Dn' *m Gn = Gn *m Dn.
Proof.
by rewrite thinmx0 flatmx0.
Qed.


Lemma D0GnG0D0':D0' *m Gn = G0 *m D0.
Proof.
by rewrite !thinmx0.
Qed.



Lemma G_ccmorphism_m:
   forall i:Z, (diff D'cc i) *m (G_m i) = (G_m (i+1)) *m (diff Dcc i).
Proof.
move=> i.
case:i.
  rewrite /=; exact D1G0G1D1'.
move=> p.
case :p.
  move=> p.
  case:p; try by move=>p; rewrite /=; exact DnGnGnDn'.
  rewrite /=; exact DnGnGnDn'.
  move=>p.
  case:p; try by move=> p; rewrite /=; exact DnGnGnDn'.
  rewrite /=; exact D3G2GnD3'.
  rewrite /=; exact  D2G1G2D2'.
move=> p.
case:p; try by move=>p; rewrite /=; exact DnGnGnDn'.
rewrite /=; exact D0GnG0D0'. 
Qed.


Definition G_m_m:= Build_FGChain_Complex_Morphism (A:=D'cc)(B:=Dcc) G_ccmorphism_m.


(* Homotopy operator*)

Variable (H2 : 'M[K]_(c2,0)).
Variable (H00 : 'M[K]_(0,c0)).
Variable (Hn : 'M[K]_(0,0)).


Definition H_m:= fun i:Z => match i as
 z return ((fun z0 : Z => 'M_(m_m z0, m_m (z0+1))) z) with
|(-1)%Z => H00
|0%Z => H0
|1%Z => H1
|2%Z => H2
|3%Z => Hn
|_ => Hn
end.

Definition H_m_m:= Build_FGHomotopy_operator (A:=Dcc) H_m.



(*Reduction*)
(*Properties of a reduction for all the dimensions*)
Lemma GF_m:  forall i:Z, (G_m i) *m (F_m i) = 1%:M.
Proof.
move=> i.
case:i.
  rewrite /=; exact ax1_0.
  move=> p.
  case:p.  
    move=> p.
    by rewrite /= thinmx0 flatmx0. 
    move=>p.
    case:p; try by move=>p; rewrite /= thinmx0 flatmx0. 
    by rewrite /= ax1_2.
    by rewrite /= ax1_1.
move=>p.
case:p; try by move=>p; rewrite /= thinmx0 flatmx0. 
by rewrite /= thinmx0 flatmx0.
Qed.


Lemma FGHDDH_m:  forall i:Z,  (F_m (i+1)) *m (G_m (i+1)) + 
   ((H_m (i+1)) *m (diff Dcc (i+1))) +
   ((diff Dcc i) *m (H_m i))  =  1%:M.
Proof.
move=> i.
case:i.
  rewrite /=; exact: ax2_1.
  move=> p.
  case:p. 
  move=>p.
  case:p; try by move=>p; rewrite /= !thinmx0.
  by rewrite /= !thinmx0.
  move=>p.
  case:p;try by move=>p; rewrite /= !thinmx0. 
  by rewrite /= !thinmx0.
  by rewrite /= !thinmx0 mul0mx GRing.addr0 ax2_2. 
move=> p.
case:p; try by move=>p; rewrite /= !thinmx0.
by rewrite /= !thinmx0 mul0mx GRing.addr0 ax2_0.
Qed.



Lemma HF_m : forall i:Z,  (H_m i) *m (F_m (i+1)) = 0.
Proof.
move=> i.
case:i.
  by rewrite /= ax3_0.
move=>p.
case:p.
move=>p.
case:p; try by move=>p; rewrite /=; exact HFn.
rewrite /= thinmx0.
by move=> p.
move=>p; case:p; try by move=>p; rewrite /= thinmx0.
by rewrite /= thinmx0.
by rewrite /= thinmx0.
move=> p.
case:p; try by move=>p; rewrite /= thinmx0.
by rewrite /= thinmx0.
by rewrite /= ax3_1.
move=>p.
case:p; try by move=>p; rewrite /= thinmx0.
by rewrite /= flatmx0.
Qed.


Lemma GH_m: forall i:Z, (G_m i) *m (H_m i) = 0.
Proof.
move=> i.
case:i.
  by rewrite /= ax4_0.
move=>p.
case:p.
move=>p.
case:p; try by move=>p; rewrite /= !thinmx0. 
by rewrite /= thinmx0.
move=> p.
case:p; try by move=>p; rewrite /= thinmx0.
by rewrite /= thinmx0.
by rewrite /= ax4_1. 
move=> p.
case:p; try by move=>p; rewrite /= thinmx0. 
by rewrite /= flatmx0.
Qed.


Lemma HH_m : forall i:Z, (H_m i) *m (H_m (i+1)) = 0.
Proof.
move=> i.
case:i.
  by rewrite /= ax5.
move=>p.
case:p.
move=>p.
case:p; try by move=>p; rewrite /= !thinmx0. 
by rewrite /= thinmx0.
move=> p.
case:p; try by move=>p; rewrite /= thinmx0.
by rewrite /= thinmx0.
by rewrite /= thinmx0.
move=> p.
case:p; try by move=>p; rewrite /= thinmx0. 
move=>p.
case:p; try by rewrite /= thinmx0.
by rewrite /= thinmx0 mul0mx.
by rewrite /= flatmx0.
Qed.


Definition red_FGReduction := Build_FGReduction (C:=Dcc)(D:= D'cc)(F:=F_m_m)
(G:=G_m_m) (H:=H_m_m) GF_m FGHDDH_m HF_m GH_m HH_m.


End decomposition.


End BPL.