(* Reduction definition *)

Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import finfun tuple ssralg matrix mxalgebra bigop.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section reduction_definition.

Local Open Scope ring_scope.

(* Reduction with just two non-null differentials *)

Record reduction :=
{ K  : fieldType;
 c0 : nat;
 c1 : nat;
 c2 : nat; 
 d0 : nat;
 d1 : nat; 
 d2 : nat;
 D1 : 'M[K]_(c1,c0);
 D1' : 'M[K]_(d1,d0);
 D2 : 'M[K]_(c2,c1);
 D2': 'M[K]_(d2,d1);
 F0 : 'M[K]_(c0,d0);
 F1 : 'M[K]_(c1,d1);
 F2 : 'M[K]_(c2,d2);
 G0 : 'M[K]_(d0,c0); 
 G1 : 'M[K]_(d1,c1);
 G2 : 'M[K]_(d2,c2);
 H0 : 'M[K]_(c0,c1); 
 H1 : 'M[K]_(c1,c2);
 nilpotency_d2d1 : (D2 *m D1) = 0;
 nilpotency_d2d1' : (D2' *m D1') = 0;
 F1_morph : (D2 *m F1) = (F2 *m D2');
 F0_morph : (D1 *m F0) = (F1 *m D1');
 G1_morph : (D2' *m G1) = (G2 *m D2);
 G0_morph : (D1' *m G0) = (G1 *m D1);
 ax1_0 : (G0 *m F0) = 1%:M;
 ax1_1 : (G1 *m F1) = 1%:M;
 ax1_2 : (G2 *m F2) = 1%:M;
 ax2_0 : (F0 *m G0) + (H0 *m D1) = 1%:M;
 ax2_1 : (F1 *m G1) + (H1 *m D2) + (D1 *m H0) = 1%:M;
 ax2_2 : (F2 *m G2) + (D2 *m H1) = 1%:M;
 ax3_0 : (H0 *m F1) = 0;
 ax3_1 : (H1 *m F2) = 0;
 ax4_1 : (G1 *m H1) = 0;
 ax4_0 : (G0 *m H0) = 0;
 ax5 : (H0 *m H1) = 0
}.

End reduction_definition.