Add LoadPath "CoqEAL".
Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice fintype fingroup.
Require Import finset zmodp ssralg perm rs_definitions seqmatrix1 advf_type advf_ord_triangular.
Require Import matrix seqmatrix_adds triangular triangular_rs permseq.
Require Import aux_lemmas rs_dvf rs_admissible Bool.


Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Local Open Scope ring_scope.



Section t_reorder.

Variable M: (seq (seq Z2)).
Variable m n:nat.
Hypotheses ismatrix: is_matrix m n M.



Lemma take_rows_s_t_take_rows_s (s:seq nat):
  M!=[::] -> size s = m -> t_take_rows_s s M = take_rows_s s M.
Proof.
move => HH H; move: ismatrix; rewrite /is_matrix.
move=>[H0|]; first by move:HH; rewrite H0.
move =>[H4 H5].
rewrite H4 in H.
apply : (@eq_from_nth _ nil); first by rewrite size_map size_iota size_map.
move => i H1.
rewrite (nth_map 0%N) ?nth_iota ?add0n.
rewrite (nth_map 0%N); first by done.
by move : H1; rewrite size_map size_iota H.
by move : H1; rewrite size_map size_iota.
by move : H1; rewrite size_map !size_iota.
Qed.


Lemma take_row_s_t_take_row_s (s:seq nat)(l:seq Z2) : 
  size s = size l -> t_take_row_s s l = take_row_s s l.
Proof.
move => H.
apply : (@eq_from_nth _ 0); first by rewrite size_map size_iota size_map.
move => i H1.
rewrite (nth_map 0%N) ?nth_iota ?add0n.
rewrite (nth_map 0%N); first by done.
by move : H1; rewrite size_map size_iota H.
by move : H1; rewrite size_map size_iota.
by move : H1; rewrite size_map !size_iota.
Qed.


Lemma take_columns_s_t_take_columns_s m1 n1(s:seq nat)(M1: matZ2) : 
  M1!= [::] -> is_matrix m1 n1 M1 -> size s = n1 ->
  t_take_columns_s s M1 = take_columns_s s M1.
Proof.
move=> H0 HH H.
rewrite /t_take_columns_s /take_columns_s.
apply: eq_in_map => l H1.
rewrite take_row_s_t_take_row_s; first by done.
have: exists j, ((j < m1) && ((rowseqmx M1 j) == l)).
  exists (index l M1).
  apply/andP; split; last by rewrite /rowseqmx (nth_index [::]).
  rewrite -index_mem in H1.
  move: HH; rewrite /is_matrix.
  move=>[h |]; first by move: H0; rewrite h.
  by move=>[-> _].
move=>[j Hj].
move: (andP Hj) =>[H2 H3].
rewrite -(eqP H3); move:HH; rewrite /is_matrix.
move=>[h |]; first by move: H0; rewrite h.
by move =>[H4 H5]; rewrite (H5 j H2).
Qed.


Lemma reorderM_t_reorderM (s1 s2:seq nat) : M!= [::] ->
size s1 = m -> size s2 = n -> perm_eq s1 (iota 0 (size s1)) ->
   t_reorderM s1 s2 M  = reorderM s1 s2 M.
Proof.
move=> HH H1 H2 H3.
rewrite /t_reorderM /reorderM (take_rows_s_t_take_rows_s HH H1).
move: ismatrix; rewrite /is_matrix.
move=>[h |]; first by move: HH; rewrite h.
move =>[H4 H5].
have H6: is_matrix m n (take_rows_s s1 M).
  rewrite /is_matrix; right; split; first by rewrite size_map H1.
  move=> i H; rewrite /take_rows_s /rowseqmx (nth_map 0%N); last by rewrite H1.
  rewrite -H1 in H.
  move: (mem_nth 0%N H).
  rewrite (perm_eq_mem H3) mem_iota /= add0n H1=> H6.
  by move: (H5 (nth 0%N s1 i) H6).
have H8: ((take_rows_s s1 M) != [::]).
  case H8: M=>[|m1 m2]; first by move: HH; rewrite H8.
  move:H8; rewrite /take_rows_s /=.
  case H8: s1 =>[|s11 s12]; last by done. 
  rewrite H8 /= in H1; rewrite -H1 in H4.
  move:H4; move/eqP; rewrite eq_sym; move/eqP => H4.
  by move: (size0nil H4)=> H51; move: HH; rewrite H51.
by rewrite (take_columns_s_t_take_columns_s H8 H6 H2).
Qed.


End t_reorder.










Section fill_perm_rs.


Lemma in_fill_iota n1 s:  longmn n1 s ->  ((fill s n1) =i (iota 0 n1)).
Proof.
move=> H1 i.
rewrite /fill mem_cat; apply/orP.
case H2:(i \in iota 0 n1).
  case H3: (i \in s); first by apply/orP; rewrite orb_true_l.
  by apply/orP; rewrite orb_false_l mem_filter /= H3 H2.
case H3: (i \in s).
  move: H1; rewrite /longmn /=; move/allP => H4.
  move: (H4 i H3); rewrite /= => H5.
  by move:H2; rewrite mem_iota /= add0n H5.
by apply/orP; rewrite orb_false_l mem_filter /= H3 H2.
Qed.



Lemma uniq_fill s n1: uniq s ->(uniq (fill s n1)).
Proof.
move=> H5. 
rewrite /fill cat_uniq H5.
move: (iota_uniq 0 n1)=> H3.
rewrite (filter_uniq  [pred i | i \notin s] H3) andb_true_r andb_true_l.
apply/hasPn; move =>x.
by rewrite inE mem_filter /=; move/andP =>[ -> _].
Qed.


Lemma perm_fill n1 (s:seq nat):
 longmn n1 s -> uniq s -> perm_eq (fill s n1) (iota 0 n1).
Proof.
move=> H1 H5.
move: (uniq_fill n1 H5) (iota_uniq 0 n1) (in_fill_iota H1) => H4 H3 H2.
apply: (uniq_perm_eq H4 H3 H2).
Qed.


Lemma uniq_index x1 x2 (s:seq nat): x1\in s -> uniq s -> index x1 s = index x2 s -> x1 = x2.
Proof.
elim H: s=>[|s1 s2]; first by rewrite /=.
move=> HI H1 H2.
rewrite /index /=.
case H3:(s1 == x1); first by case H4:(s1 == x2); first by rewrite -(eqP H3) -(eqP H4).
case H5:(s1 == x2 ); first by done.
case H4: s2=>[|s21 s22]; first by move:H1; rewrite H4 /= inE => h6; move:H3; rewrite (eqP h6); move/eqP.
have H6: (s2!=[::]); first by rewrite H4.
move:H2; rewrite /uniq -/uniq; move/andP =>[H7 H8]; rewrite -H4 -/index.
have h9: (x1\in s2).
  by move:H1; rewrite inE; move/orP=>[H9 | H10]; first by move:H3; rewrite (eqP H9); move/eqP.
move=> H10.
move:(eq_add_S (find (pred1 x1) s2) (find (pred1 x2) s2) H10)=> H11.
have H12: (index x1 s2 = index x2 s2 ); first by rewrite /index.
apply: (HI h9 H8 H12).
Qed.



End fill_perm_rs.



Section reorderM_ismatrix.

Variable m n:nat.
Variable M: matZ2.
Variable dvf :seq (nat*nat).
Variable dvf2 :seq (nat*nat).
Variable ords: orders.

Hypotheses vecfield: Vecfieldadm M dvf ords.
Hypotheses vecfield2: Vecfieldadm2 M dvf2 ords.
Hypotheses dvf_nonil: dvf!=[::].

Lemma M_nonil: (M!=[::]).
Proof.
case H1: M; last by done.
move: vecfield; rewrite /Vecfieldadm; move=>[H6 _].
move:H6; rewrite H1 /longmn /= all_pred0 -sizedvf_getfirst.
by case H2: dvf; first by move:dvf_nonil;rewrite H2.
Qed.

Hypotheses ismatrix: is_matrix m n M.


Lemma sizeMm: size M = m.
Proof.
move: vecfield; rewrite /Vecfieldadm; move => [H2 _].
move: ismatrix; rewrite /is_matrix; move => [H0| [H3 H21]].
  move: H2; rewrite H0 /longmn /= all_pred0.
  rewrite -aux_lemmas.sizedvf_getfirst.
  move/eqP => H21. move: (size0nil H21).
  by move=> H3; move: dvf_nonil; rewrite H3.
by rewrite H3.
Qed.


Lemma sizeheadMn:  (size (nth [::] M 0)) = n.
Proof.
move: vecfield; rewrite /Vecfieldadm; move=> [_ [H1 _]].
move: ismatrix; rewrite /is_matrix; move => [H0| [H3 H21]].
  move: H1; rewrite H0 /longmn /= all_pred0.
  rewrite -aux_lemmas.sizedvf_getsnd.
  move/eqP => H21. move: (size0nil H21).
  by move=> H3; move: dvf_nonil; rewrite H3.
have H4: (0 < m).
  case H5: m; last by done.
  move: H3; rewrite H5.
  move/eqP; rewrite eq_sym. move/nilP => H7.
  move: vecfield; rewrite /Vecfieldadm; move =>[H6 _].
  move: H6; rewrite H7 /longmn /= all_pred0.
  rewrite -aux_lemmas.sizedvf_getfirst.
  move/eqP => H4. move: (size0nil H4).
  by move=> H3; move: dvf_nonil; rewrite H3.
by move: (H21 0%N H4) => ->.
Qed.


Lemma dvf_nonil_M_nonil: M = [::]->  dvf =[::].
Proof.
move=> H2.
move: vecfield; rewrite /Vecfieldadm H2; move =>[H3 _].
move:H3; rewrite /longmn /= all_pred0 -sizedvf_getfirst.
by move/eqP=> H3; move: (size0nil H3). 
Qed.



Lemma Om: 0 < m.
Proof.
case H5: m; last by done.
move:vecfield; rewrite /Vecfieldadm; move=>[H6 _].
move:H6; rewrite sizeMm H5 /longmn /= all_pred0.
by case hh:dvf; first by move:dvf_nonil; rewrite hh.
Qed.


Lemma On: 0 < n.
Proof.
case H5: n; last by done.
move:vecfield; rewrite /Vecfieldadm; move=>[H61 [H6 _]];move:H6.
move: ismatrix; rewrite /is_matrix; move => [H2|H2].
  by move: (dvf_nonil_M_nonil H2) dvf_nonil => ->.
move: H2; move=>[H21 H22].
move: Om => H51.
move: (H22 0%N H51) => ->.
rewrite H5 /longmn /= all_pred0.
by case hh:dvf; first by move:dvf_nonil; rewrite hh.
Qed.



Lemma M_nonil_seqmx_nonil: (seqmx_of_mx (mx_of_seqmx m n M))!=[::].
Proof.
case H: M.
  by move: (dvf_nonil_M_nonil H) dvf_nonil => ->.
have : (M!=[::]); first by rewrite H.
rewrite -H.
move: ismatrix; rewrite /is_matrix; move => [->|H2]; first by done.
by rewrite seqmx_of_mx_mx_of_seqmx.
Qed.



Lemma size_row_ulsubseqmx1 m1 n1 j (M1: matZ2): M1!=[::] ->
  is_matrix m n M1 ->
  j < m1 -> m1 <= (size M1)  ->  n1 <= n -> 
  size (rowseqmx (ulsubseqmx m1 n1 M1) j) = n1.
Proof.
move => H0 hh.
have: (is_matrix m n M1); first by done.
rewrite /is_matrix. move => [h0|]; first by move: H0; rewrite h0.
move=> [H1 H2] h3 H4.
rewrite /rowseqmx /ulsubseqmx (nth_map [::]); last first.
  rewrite size_take. 
  case H6:(m1 < size M1); first by done.
  rewrite -H1.
  move: H4;rewrite leq_eqVlt; move/orP=>[H51|H51]; first by rewrite H1 -(eqP H51).
  by move:H51; rewrite H6.
rewrite size_take nth_take; last by done.
case H41:(n1 < size (nth [::] M1 j)); first by done.
move: H4;rewrite leq_eqVlt; move/orP=>[H51|H51].
  rewrite leq_eqVlt; move/orP=>[H6|H6].
    rewrite (eqP H6); move: h3; rewrite (eqP H51) -H1 => H7.
    by move: (H2 j H7); rewrite /=.
  move: h3; rewrite (eqP H51) -H1 => H7.
  move: (H2 j H7); rewrite /rowseqmx => H8.
  rewrite H8 in H41. 
  by move:H6; rewrite H41.
rewrite leq_eqVlt; move/orP=>[H6|H6].
  move: (ltn_trans h3 H51).
  by rewrite -H1=>  H7; move: (H2 j H7); rewrite (eqP H6).
move: (ltn_trans h3 H51).
rewrite -H1=>  H7; move: (H2 j H7); rewrite /rowseqmx; move=> H8.
by move: H41; rewrite H8 H6.
Qed.


Lemma uniq_filter s1 s2 nn:
  uniq (s1::s2) -> size (filter [pred x | (x == s1) && (x \in s2)] (iota 0 nn)) = 0%N.
Proof.
elim H: nn=> [|nn1]; first by rewrite /=.
move=> H2 H1.
rewrite -addn1 iota_add filter_cat size_cat (H2 H1) add0n /iota add0n /=.
case H3:((nn1 == s1) && (nn1 \in s2)); last by done.
move: (andP H3)=>[H31 H32]; move:H1; rewrite /=.
by move: H32; rewrite (eqP H31) => ->.
Qed.


Lemma count_pred nn s: uniq s -> 
  (count [pred i \in s] (iota 0 nn) = count [pred i| i < nn] s).
Proof.
elim H: s =>[|s1 s2]; first by rewrite {2}/count count_pred0.
move=> H1 HH.
rewrite /= -/count -H1; last by move:HH; rewrite /=; move/andP=>[_ ->].
have ->: (count [pred i \in s1 :: s2] (iota 0 nn))%N =  
   (((count (pred1 s1) (iota 0 nn))%N) + ((count [pred i \in s2] (iota 0 nn))%N))%N; last first.
  move: ((iota_uniq 0 nn))=> h2.
  by apply/eqP; rewrite eqn_addr (count_uniq_mem); first by rewrite mem_iota add0n.
rewrite -count_predUI /predU.
have ->:  (count (predI (pred1 s1) [pred i \in s2]) (iota 0 nn))%N = 0%N; last first.
  by have ->: [pred i \in s1 :: s2] = [pred x | (x == s1) || (x \in s2)]; last by rewrite addn0.
rewrite count_filter /predI /=.
move: (iota_uniq 0 nn)=> H2.
by apply: (uniq_filter).
Qed.



Lemma longmn_fill_size nn (s:seq nat):  longmn nn s -> 
  uniq s -> nn = size (fill s nn).
Proof.
move=> H1 H2.
rewrite /fill !size_cat /=. 
have H3:(size s = size (filter [pred i | i \in s] (iota 0 nn))); last first.
  by rewrite H3 -!count_filter /= count_predC size_iota.
move:H1; rewrite /longmn /= all_count -count_filter => H3.
by rewrite count_pred; first by rewrite (eqP H3).
Qed.


Lemma sizeM_reorderM : size M = size (reorderM_dvf dvf M).
Proof.
case H: dvf =>[|v1 v2]; first by rewrite /reorderM_dvf /reorderM /fill !cat0s /take_columns_s /take_rows_s /take_row_s !size_map /= size_iota.
rewrite /reorderM_dvf /reorderM /take_columns_s !size_map.
move: vecfield; rewrite /Vecfieldadm.
move => [H1 H2]; move:H2 =>[H2 H3]; move:H3=>[H3 H4].
by rewrite {1}(longmn_fill_size H1 H3) H.
Qed.


Lemma sizeM_reorderM2 dvf3 (H1 : longmn (size M) (getfirstElemseq dvf3)) (H2: uniq (getfirstElemseq dvf3)) : size M = size (reorderM_dvf dvf3 M).
Proof.
case H: dvf3 =>[|v1 v2]; first by rewrite /reorderM_dvf /reorderM /fill !cat0s /take_columns_s /take_rows_s /take_row_s !size_map /= size_iota.
rewrite /reorderM_dvf /reorderM /take_columns_s !size_map.
by rewrite {1}(longmn_fill_size H1 H2) H.
Qed.


Lemma is_matrixReorderM : M!= [::] ->
  is_matrix m n (reorderM_dvf dvf M).
Proof.
move: (size_dvf_reorderM dvf M) => HH H1.
rewrite /is_matrix.
right; split.
  move: ismatrix; rewrite /is_matrix.
  move => [h0|]; first by move: H1; rewrite h0.
  by move=>  [H2 _]; rewrite -sizeM_reorderM.
move: ismatrix; rewrite /is_matrix.
move => [h0|]; first by move: H1; rewrite h0.
move=>  [H2 H3] i im.
move: (H3 i im).
rewrite /reorderM_dvf /reorderM /rowseqmx => H4.
rewrite /take_columns_s /=  (nth_map [::]); last first.
  rewrite /take_rows_s size_map. 
  move: vecfield; rewrite /Vecfieldadm.
  move => [H11 H21]; move:H21 =>[H21 H31]; move:H31=>[H31 H41].
  by rewrite -(longmn_fill_size H11 H31) -H2.
rewrite /take_rows_s /take_row_s size_map.
move: vecfield; rewrite /Vecfieldadm.
move => [_ H21]; move:H21 =>[H21 H31]; move:H31=>[_ H41]; move: H41=> [H41 H5].
rewrite -(longmn_fill_size H21 H41).
move: Om => H6.
by move: (H3 0%N H6). 
Qed.


(*size_dvf_reorderM:   the size of dvf is lower equal the number of rows *)
(*the size of dvf is lower equal the number of columns*)
Lemma size_dvf_lower_ncolumns: size dvf <= n.
Proof.
move: (sizedvf_getsnd dvf)=> H1.
move: vecfield; rewrite /Vecfieldadm.
move=>[H2 [H3 [H4 [H5 _]]]].
move: (perm_fill H3 H5) => H6.
move: (perm_eq_size H6).
rewrite size_iota; move: ismatrix; rewrite /is_matrix.
move=>[h0|]; first by move:dvf_nonil; rewrite (dvf_nonil_M_nonil).
move=>[H7 H8].
move:(H8 0%N Om)=> ->.
by rewrite H1 /fill size_cat => <-; rewrite leq_addr.
Qed.



Lemma size_rowseqmx_sizedvf i: 
  i < size dvf- 1 +1 ->
  size (rowseqmx (ulsubseqmx (size dvf - 1 + 1) (size dvf - 1 + 1) 
  (reorderM_dvf dvf M)) i) = (size dvf - 1 + 1)%N.
Proof.
move:(M_nonil) => HH H4.
move: (is_matrixReorderM )=> H.
move: (size_dvf_reorderM dvf M) => H2.
have H6: (size dvf -1 +1) <= size (reorderM_dvf dvf M). 
  rewrite subn1 addn1 /=.
  case H7: dvf=>[|v1 v2]; first by move: dvf_nonil; rewrite H7.
  by move:H2; rewrite H7 /=.
have H8: (size dvf - 1 + 1 <= n); last first.
  have H9:((reorderM_dvf dvf M)!=[::]).
    case H9: (reorderM_dvf dvf M); last by done.
    move: (sizeM_reorderM); rewrite H9 /= => H10.
    by move: (size0nil H10) HH => ->.
  move:(@size_row_ulsubseqmx1 (size dvf - 1 + 1) 
  (size dvf - 1 + 1) i (reorderM_dvf dvf M) H9 (H HH) H4 H6 H8).
  have H3:((size dvf - 1 + 1)%N = size dvf).
    case Hh: dvf =>[|v1 v2]; first by move:dvf_nonil; rewrite Hh.
    by rewrite /= subn1 addn1 /=. 
  by rewrite H3; move:H; rewrite /is_matrix; move=> h ->.
case H1:(size dvf)=>[|n1]; first by move: dvf_nonil; rewrite (size0nil H1). 
rewrite subn1 addn1 /=.
by move: (size_dvf_lower_ncolumns); rewrite H1.
Qed.


End reorderM_ismatrix.



Section reorderM_equivalence.

Variable s s1: seqnat.
Variable m n:nat.
Hypothesis H : perm_eq s (iota 0 m).
Hypothesis H1 : perm_eq s1 (iota 0 n).


Lemma reorderM_E (M1:'M[Z2]_(m,n)) :
  (seqmx_of_mx M1) != [::] -> size s = m -> size s1 = n -> 
   reorderM s s1 (seqmx_of_mx M1) = (seqmx_of_mx (reorder_mx (perm_of_seq H s) (perm_of_seq H1 s1) M1)).
Proof.
move=> h0 H11 H2.
have H4: (is_matrix m n (seqmx_of_mx M1)).
  rewrite /is_matrix.
  right; split; first by rewrite size_map size_enum_ord.
  by move=> i H4; rewrite size_row_seqmx.
have H3: ( perm_eq s (iota 0 (size s))); first by rewrite H11. 
by rewrite -(reorderM_t_reorderM H4 h0 H11 H2 H3) t_reorderM_E.
Qed.


Lemma reorderM2_E (M1:matZ2) : M1!= [::] ->
 size s = m -> size s1 = n -> perm_eq s (iota 0 (size s)) ->
 is_matrix m n M1 ->
  reorder_mx (perm_of_seq H s) (perm_of_seq H1 s1) (mx_of_seqmx m n M1) = mx_of_seqmx m n (reorderM s s1 M1).
Proof.
move => H0 H11 H3 H4.
rewrite /is_matrix. move =>[H2|H2]; first by move:H0; rewrite H2.
rewrite -{2}(seqmx_of_mx_mx_of_seqmx H2) reorderM_E ?mx_of_seqmx_seqmx_of_mx; try by done.
by rewrite (seqmx_of_mx_mx_of_seqmx H2).
Qed.



End reorderM_equivalence.









